from _balder.balder_session import BalderSession
from _balder.exceptions import VDeviceOverwritingError, BalderException

from tests.test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0OverwriteIllegallyBetsceset(Base0EnvtesterClass):
    """
    This testcase is a modified version of the basic ENVTESTER environment. It is limited to the `ScenarioA` and the
    `SetupA` and uses a feature `FeatureOfRelevance` for testing the correct overwriting of VDevices BETWEEN the
    scenario-feature and the setup-feature.

    This test defines a new feature `FeatureOfRelevanceLvl1` with two vDevices, which is directly instantiated in the
    scenario. This feature will be overwritten by a direct child `FeatureOfRelevanceLvl2` that overwrites both vDevices.
    Instead, of made as in the test `test_0_overwrite_correctly_betsceset`, not all vDevices are being overwritten here.
    The setup device instantiates a child class (`SetupFeatureOfRelevanceLvl3`) of it.

    The test should terminate successful and no error should occur!
    """

    @property
    def expected_data(self) -> tuple:
        return ()

    @property
    def expected_exit_code(self) -> int:
        return 4

    @staticmethod
    def handle_balder_exception(exc: BalderException):
        assert isinstance(exc, VDeviceOverwritingError), 'unexpected exception type'
        assert exc.args[0] == "missing overwriting of parent VDevice class `FeatureOfRelevanceLvl1.VDeviceWithII` in " \
                              "feature class `FeatureOfRelevanceLvl2` - if you overwrite one or more VDevice(s) you " \
                              "have to overwrite all!"

    @staticmethod
    def validate_finished_session(session: BalderSession):
        assert session.executor_tree is None, "test session does not terminates before collector work was done"



from typing import Union
from _balder.testresult import ResultState
from _balder.balder_session import BalderSession

from tests.test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0TreecheckFixtSetupaSessionConstruction(Base0EnvtesterClass):
    """
    This testcase executes the basic envtester and forces an error on a specific given position. The test checks if the
    system behaviour is as expected.

    file: ``setups/setup_a.py``
    class: ``SetupA``
    method/function: ``fixture_session``
    part: ``construction``
    """

    @property
    def cmd_args(self):
        return [
            '--test-error-file', 'setups/setup_a.py',
            '--test-error-cls', 'SetupA',
            '--test-error-meth', 'fixture_session',
            '--test-error-part', 'construction',
        ]

    @property
    def expected_exit_code(self):
        return 1

    @property
    def expected_data(self):
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupB.fixture_session
            {"cls": "SetupB", "meth": "fixture_session", "part": "construction"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_session
            {"cls": "SetupA", "meth": "fixture_session", "part": "construction"},
            # {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            # {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # stops execution here (only valid if the ScenarioB fixture was called first)
            # {"cls": "SetupA", "meth": "fixture_session", "part": "teardown"},

            # FIXTURE-TEARDOWN: SetupA.fixture_session
            {"cls": "SetupB", "meth": "fixture_session", "part": "teardown"},
            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
            # FIXTURE-TEARDOWN: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "teardown"}
        )

    @property
    def expected_data_alternative(self) -> Union[tuple, None]:
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "construction"},
            # FIXTURE-CONSTRUCTION: SetupA.fixture_session
            {"cls": "SetupA", "meth": "fixture_session", "part": "construction"},

            # {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
            # {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
            # stops execution here (only valid if the ScenarioB fixture was called first)
            # {"cls": "SetupA", "meth": "fixture_session", "part": "teardown"},

            # FIXTURE-TEARDOWN: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "teardown"},
        )

    @staticmethod
    def validate_finished_session(session: BalderSession):
        # check result states everywhere
        assert session.executor_tree.executor_result == ResultState.ERROR, "test session does not terminates with ERROR"

        assert session.executor_tree.construct_result.result == ResultState.ERROR, \
            "global executor tree construct part does not set ResultState.ERROR"
        assert session.executor_tree.body_result.result == ResultState.NOT_RUN, \
            "global executor tree body part does not set ResultState.NOT_RUN"
        assert session.executor_tree.teardown_result.result == ResultState.SUCCESS, \
            "global executor tree teardown part does not set ResultState.SUCCESS"
        for cur_setup_executor in session.executor_tree.get_setup_executors():
            assert cur_setup_executor.executor_result == ResultState.NOT_RUN, \
                "the setup executor does not have result NOT_RUN"

            assert cur_setup_executor.construct_result.result == ResultState.NOT_RUN
            assert cur_setup_executor.body_result.result == ResultState.NOT_RUN
            assert cur_setup_executor.teardown_result.result == ResultState.NOT_RUN

            for cur_scenario_executor in cur_setup_executor.get_scenario_executors():
                assert cur_scenario_executor.executor_result == ResultState.NOT_RUN, \
                    "the scenario executor does not have result NOT_RUN"

                assert cur_scenario_executor.construct_result.result == ResultState.NOT_RUN
                assert cur_scenario_executor.body_result.result == ResultState.NOT_RUN
                assert cur_scenario_executor.teardown_result.result == ResultState.NOT_RUN

                for cur_variation_executor in cur_scenario_executor.get_variation_executors():
                    assert cur_variation_executor.executor_result == ResultState.NOT_RUN, \
                        "the variation executor does not have result NOT_RUN"

                    assert cur_variation_executor.construct_result.result == ResultState.NOT_RUN
                    assert cur_variation_executor.body_result.result == ResultState.NOT_RUN
                    assert cur_variation_executor.teardown_result.result == ResultState.NOT_RUN

                    for cur_testcase_executor in cur_variation_executor.get_testcase_executors():
                        assert cur_testcase_executor.executor_result == ResultState.NOT_RUN, \
                            "the testcase executor does not have result NOT_RUN"

                        assert cur_testcase_executor.construct_result.result == ResultState.NOT_RUN
                        assert cur_testcase_executor.body_result.result == ResultState.NOT_RUN
                        assert cur_testcase_executor.teardown_result.result == ResultState.NOT_RUN



import balder
import logging
from ..balderglob import RuntimeObserver
from .scenario_a_parent import ScenarioAParent

logger = logging.getLogger(__file__)


class ScenarioAChild(ScenarioAParent):
    """This is the CHILD CLASS scenario of category A"""

    class ScenarioDevice1(ScenarioAParent.ScenarioDevice1):
        pass

    # missing overwritten `ScenarioDevice2`

    # this is our overwritten test case
    def test_a_2(self):
        RuntimeObserver.add_entry(__file__, ScenarioAChild, ScenarioAChild.test_a_2, category="testcase",
                                  msg=f"execute Test `{ScenarioAChild.test_a_2.__qualname__}`")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

    @balder.fixture(level="session")
    def fixture_session(self):
        RuntimeObserver.add_entry(__file__, ScenarioAChild, ScenarioAChild.fixture_session, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioAChild, ScenarioAChild.fixture_session, category="fixture",
                                  part="teardown", msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

    @balder.fixture(level="setup")
    def fixture_setup(self):
        RuntimeObserver.add_entry(__file__, ScenarioAChild, ScenarioAChild.fixture_setup, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioAChild, ScenarioAChild.fixture_setup, category="fixture",
                                  part="teardown", msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

    @balder.fixture(level="scenario")
    def fixture_scenario(self):
        RuntimeObserver.add_entry(__file__, ScenarioAChild, ScenarioAChild.fixture_scenario, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioAChild, ScenarioAChild.fixture_scenario, category="fixture",
                                  part="teardown", msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

    @balder.fixture(level="variation")
    def fixture_variation(self):
        RuntimeObserver.add_entry(__file__, ScenarioAChild, ScenarioAChild.fixture_variation, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioAChild, ScenarioAChild.fixture_variation, category="fixture",
                                  part="teardown", msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

    @balder.fixture(level="testcase")
    def fixture_testcase(self):
        RuntimeObserver.add_entry(__file__, ScenarioAChild, ScenarioAChild.fixture_testcase, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioAChild, ScenarioAChild.fixture_testcase, category="fixture",
                                  part="teardown", msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()



import balder
from .between_meth_var_feature import BetweenMethVarFeature
from ...lib.connections import AConnection, BConnection, CConnection
from ...balderglob import RuntimeObserver


@balder.for_vdevice("VDeviceI", with_connections=balder.Connection.based_on(AConnection, BConnection, CConnection))
@balder.for_vdevice("VDeviceII", with_connections=balder.Connection.based_on(AConnection, BConnection))
@balder.for_vdevice("VDeviceIII", with_connections=balder.Connection.based_on(BConnection, (AConnection, BConnection)))
class SetupMethVarFeature(BetweenMethVarFeature):
    """This is the feature class that is directly being used in the setup."""
    class VDeviceI(BetweenMethVarFeature.VDeviceI):
        pass

    class VDeviceII(BetweenMethVarFeature.VDeviceII):
        pass

    class VDeviceIII(BetweenMethVarFeature.VDeviceIII):
        pass

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupMethVarFeature, SetupMethVarFeature.do_something,
            "enter `SetupMethVarFeature.do_something`",
            category="feature")

    @balder.for_vdevice("VDeviceI", with_connections=AConnection)
    def do_something_as_var(self):
        super().do_something_as_var()
        RuntimeObserver.add_entry(
            __file__, SetupMethVarFeature, SetupMethVarFeature.do_something_as_var,
            "MethodVariation<VDeviceI|AConnection><SetupMethVarFeature.do_something_as_var>", category="feature")

    @balder.for_vdevice("VDeviceI", with_connections=BConnection)
    def do_something_as_var(self):
        super().do_something_as_var()
        RuntimeObserver.add_entry(
            __file__, SetupMethVarFeature, SetupMethVarFeature.do_something_as_var,
            "MethodVariation<VDeviceI|BConnection><SetupMethVarFeature.do_something_as_var>", category="feature")

    @balder.for_vdevice("VDeviceI", with_connections=CConnection)
    def do_something_as_var(self):
        super().do_something_as_var()
        RuntimeObserver.add_entry(
            __file__, SetupMethVarFeature, SetupMethVarFeature.do_something_as_var,
            "MethodVariation<VDeviceI|CConnection><SetupMethVarFeature.do_something_as_var>", category="feature")

    @balder.for_vdevice("VDeviceII", with_connections=balder.Connection.based_on(AConnection, BConnection))
    def do_something_as_var(self):
        super().do_something_as_var()
        RuntimeObserver.add_entry(
            __file__, SetupMethVarFeature, SetupMethVarFeature.do_something_as_var,
            "MethodVariation<VDeviceII|AConnection+OR+BConnection><SetupMethVarFeature.do_something_as_var>",
            category="feature")

    @balder.for_vdevice("VDeviceIII", with_connections=BConnection)
    def do_something_as_var(self):
        super().do_something_as_var()
        RuntimeObserver.add_entry(
            __file__, SetupMethVarFeature, SetupMethVarFeature.do_something_as_var,
            "MethodVariation<VDeviceIII|BConnection><SetupMethVarFeature.do_something_as_var>",
            category="feature")

    @balder.for_vdevice("VDeviceIII", with_connections=balder.Connection.based_on((AConnection, BConnection)))
    def do_something_as_var(self):
        super().do_something_as_var()
        RuntimeObserver.add_entry(
            __file__, SetupMethVarFeature, SetupMethVarFeature.do_something_as_var,
            "MethodVariation<VDeviceIII|AConnection+AND+BConnection><SetupMethVarFeature.do_something_as_var>",
            category="feature")



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_2_msg_transmission(balder_working_dir):
    """
    This testcase executes the basic example for transmitting a message between two devices. It checks that the
    Balder process terminates with exit code 0.

    This environment provides one single scenario that consists of one Device which implements the feature
    :class:`SendMessageFeature` and one Device which implements the feature :class:`ReceiveMessageFeature`. It holds
    two different feature implementations, while the
    :class:`SendOverLocalVarFeature`/:class:`ReceiveOverLocalVarFeature` transmits the message over a local internal
    variable inside the receive-feature and the :class:`SendOverQueueFeature`/:class:`ReceiveOverQueueFeature` uses a
    queue between these two features to transmit the message.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()
    print("\n")

    assert session.executor_tree.executor_result == ResultState.SUCCESS, \
        "test session does not terminates with success"



import balder
import logging
from ..lib.features import FeatureI, FeatureII
from ..lib.connections import AConnection
from ..lib.utils import FixtureReturn
from ..balderglob import RuntimeObserver

logger = logging.getLogger(__file__)


class ScenarioA(balder.Scenario):
    """This is the scenario of category A"""

    class ScenarioDevice1(balder.Device):
        i = FeatureI()

    @balder.connect(ScenarioDevice1, over_connection=AConnection)
    class ScenarioDevice2(balder.Device):
        ii = FeatureII()

    def test_a_1(self, fixture_session, fixture_setup, fixture_scenario, fixture_variation, fixture_testcase):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.test_a_1, category="testcase",
                                  msg=f"execute Test `{ScenarioA.test_a_1.__qualname__}`")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        assert isinstance(self, ScenarioA)
        assert fixture_session == FixtureReturn.SCENARIO_A_SESSION
        assert fixture_setup == FixtureReturn.SCENARIO_A_SETUP
        assert fixture_scenario == FixtureReturn.SCENARIO_A_SCENARIO
        assert fixture_variation == FixtureReturn.SCENARIO_A_VARIATION
        assert fixture_testcase == FixtureReturn.SCENARIO_A_TESTCASE


    def test_a_2(self, fixture_session, fixture_setup, fixture_scenario, fixture_variation, fixture_testcase):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.test_a_2, category="testcase",
                                  msg=f"execute Test `{ScenarioA.test_a_2.__qualname__}`")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        assert isinstance(self, ScenarioA)
        assert fixture_session == FixtureReturn.SCENARIO_A_SESSION
        assert fixture_setup == FixtureReturn.SCENARIO_A_SETUP
        assert fixture_scenario == FixtureReturn.SCENARIO_A_SCENARIO
        assert fixture_variation == FixtureReturn.SCENARIO_A_VARIATION
        assert fixture_testcase == FixtureReturn.SCENARIO_A_TESTCASE

    @balder.fixture(level="session")
    def fixture_session(self):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_session, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        assert isinstance(self, ScenarioA)

        yield FixtureReturn.SCENARIO_A_SESSION

        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_session, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

    @balder.fixture(level="setup")
    def fixture_setup(self, fixture_session, fixture_setup):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_setup, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        assert isinstance(self, ScenarioA)
        assert fixture_session == FixtureReturn.SCENARIO_A_SESSION
        assert fixture_setup == FixtureReturn.SETUP_A_SETUP

        yield FixtureReturn.SCENARIO_A_SETUP

        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_setup, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

    @balder.fixture(level="scenario")
    def fixture_scenario(self, fixture_session, fixture_setup, fixture_scenario):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_scenario, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        assert isinstance(self, ScenarioA)
        assert fixture_session == FixtureReturn.SCENARIO_A_SESSION
        assert fixture_setup == FixtureReturn.SCENARIO_A_SETUP
        assert fixture_scenario == FixtureReturn.SETUP_A_SCENARIO

        yield FixtureReturn.SCENARIO_A_SCENARIO

        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_scenario, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

    @balder.fixture(level="variation")
    def fixture_variation(self, fixture_session, fixture_setup, fixture_scenario, fixture_variation):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_variation, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        assert isinstance(self, ScenarioA)
        assert fixture_session == FixtureReturn.SCENARIO_A_SESSION
        assert fixture_setup == FixtureReturn.SCENARIO_A_SETUP
        assert fixture_scenario == FixtureReturn.SCENARIO_A_SCENARIO
        assert fixture_variation == FixtureReturn.SETUP_A_VARIATION

        yield FixtureReturn.SCENARIO_A_VARIATION

        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_variation, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

    @balder.fixture(level="testcase")
    def fixture_testcase(self, fixture_session, fixture_setup, fixture_scenario, fixture_variation, fixture_testcase):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_testcase, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        assert isinstance(self, ScenarioA)
        assert fixture_session == FixtureReturn.SCENARIO_A_SESSION
        assert fixture_setup == FixtureReturn.SCENARIO_A_SETUP
        assert fixture_scenario == FixtureReturn.SCENARIO_A_SCENARIO
        assert fixture_variation == FixtureReturn.SCENARIO_A_VARIATION
        assert fixture_testcase == FixtureReturn.SETUP_A_TESTCASE

        yield FixtureReturn.SCENARIO_A_TESTCASE

        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_testcase, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()
